function ampCount(selector) {
    const len = selector.length;
    let cnt = 0;
    for (let i = 0; i < len; ++i) {
        if (selector[i] === '&')
            ++cnt;
    }
    return cnt;
}
/**
 * selector must includes '&'
 */
function resolveSelectorWithAmp(amp, selector) {
    const nextAmp = [];
    selector.split(seperatorRegex).forEach(selectorPart => {
        let round = ampCount(selectorPart);
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (!round) {
            amp.forEach(ampPart => {
                nextAmp.push((ampPart + ' ' + selectorPart).trim());
            });
            return;
        }
        let partialNextAmp = [
            selectorPart
        ];
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        while (round--) {
            const nextPartialNextAmp = [];
            partialNextAmp.forEach(selectorItr => {
                amp.forEach(ampPart => {
                    nextPartialNextAmp.push(selectorItr.replace('&', ampPart));
                });
            });
            partialNextAmp = nextPartialNextAmp;
        }
        partialNextAmp.forEach(part => nextAmp.push(part));
    });
    return nextAmp;
}
/**
 * selector mustn't includes '&'
 */
function resolveSelector(amp, selector) {
    const result = [];
    selector.split(seperatorRegex).forEach(selectorPart => {
        amp.forEach(ampPart => {
            result.push((ampPart + ' ' + selectorPart).trim());
        });
    });
    return result;
}
const seperatorRegex = /,(?![^(]*\))/;
const extraSpaceRegex = /\s+/g;
function parseSelectorPath(selectorPaths) {
    let amp = [''];
    selectorPaths.forEach(selector => {
        // eslint-disable-next-line
        selector = selector && selector.trim();
        if (
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        !selector) {
            /**
             * if it's a empty selector, do nothing
             */
            return;
        }
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (selector.includes('&')) {
            amp = resolveSelectorWithAmp(amp, selector);
        }
        else {
            amp = resolveSelector(amp, selector);
        }
    });
    return amp.join(', ').trim().replace(extraSpaceRegex, ' ');
}

const kebabRegex = /[A-Z\u00C0-\u00D6\u00D8-\u00DE]/g;
function kebabCase(pattern) {
    return pattern.replace(kebabRegex, match => '-' + match.toLowerCase());
}
/** TODO: refine it to solve nested object */
function upwrapProperty(prop, indent = '  ') {
    if (typeof prop === 'object' && prop !== null) {
        return (' {\n' +
            Object.entries(prop).map(v => {
                return indent + `  ${kebabCase(v[0])}: ${v[1]};`;
            }).join('\n') +
            '\n' + indent + '}');
    }
    return `: ${String(prop)};`;
}
/** unwrap properties */
function upwrapProperties(props, instance, params) {
    if (typeof props === 'function') {
        return props({
            context: instance.context,
            props: params
        });
    }
    return props;
}
function createStyle(selector, props, instance, params) {
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (!props)
        return '';
    // eslint-disable-next-line
    const unwrappedProps = upwrapProperties(props, instance, params);
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (!unwrappedProps)
        return '';
    const propertyNames = Object.keys(unwrappedProps);
    if (propertyNames.length === 0) {
        if (instance.config.preserveEmptyBlock)
            return selector + ' {\n}';
        return '';
    }
    const statements = [
        selector + ' {'
    ];
    propertyNames.forEach(propertyName => {
        const property = unwrappedProps[propertyName];
        propertyName = kebabCase(propertyName);
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (property !== null && property !== undefined) {
            statements.push(`  ${propertyName}${upwrapProperty(property)}`);
        }
    });
    statements.push('}');
    return statements.join('\n');
}
function loopCNodeListWithCallback(children, options, callback) {
    /* istanbul ignore if */
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (!children)
        return;
    children.forEach(child => {
        if (Array.isArray(child)) {
            loopCNodeListWithCallback(child, options, callback);
        }
        else if (typeof child === 'function') {
            const grandChildren = child(options);
            if (Array.isArray(grandChildren)) {
                loopCNodeListWithCallback(grandChildren, options, callback);
                // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
            }
            else if (grandChildren) {
                callback(grandChildren);
            }
            // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        }
        else if (child) {
            callback(child);
        }
    });
}
function traverseCNode(node, selectorPaths, styles, instance, params) {
    const $ = node.$;
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (!$ || typeof $ === 'string') {
        // as a string selector
        selectorPaths.push($);
    }
    else if (typeof $ === 'function') {
        // as a lazy selector
        selectorPaths.push($({
            context: instance.context,
            props: params
        }));
    }
    else { // as a option selector
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if ($.before)
            $.before(instance.context);
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (!$.$ || typeof $.$ === 'string') {
            selectorPaths.push($.$);
            // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        }
        else /* istanbul ignore else */ if ($.$) {
            selectorPaths.push($.$({
                context: instance.context,
                props: params
            }));
        }
    }
    const selector = parseSelectorPath(selectorPaths);
    const style = createStyle(selector, node.props, instance, params);
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (style.length)
        styles.push(style);
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (node.children) {
        loopCNodeListWithCallback(node.children, {
            context: instance.context,
            props: params
        }, childNode => {
            traverseCNode(childNode, selectorPaths, styles, instance, params);
        });
    }
    selectorPaths.pop();
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (typeof $ === 'object' && $ && $.after)
        $.after(instance.context);
}
function render(node, instance, props) {
    const styles = [];
    traverseCNode(node, [], styles, instance, props);
    return styles.join('\n\n');
}

function removeElement(el) {
    /* istanbul ignore if */
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (!el)
        return;
    const parentElement = el.parentElement;
    /* istanbul ignore else */
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (parentElement)
        parentElement.removeChild(el);
}
function queryElement(id) {
    return document.querySelector(`style[cssr-id="${String(id)}"]`);
}
function createElement(id) {
    const el = document.createElement('style');
    el.setAttribute('cssr-id', String(id));
    return el;
}

function getCountOfElement(el) {
    const count = el.getAttribute('mount-count');
    if (count === null)
        return 0;
    return Number(count);
}
function setCountOfElement(el, count) {
    if (count === null) {
        el.removeAttribute('mount-count');
    }
    else {
        el.setAttribute('mount-count', String(count));
    }
}
function unmount(intance, node, target, count) {
    const els = node.els;
    if (target === undefined) {
        els.forEach(removeElement);
        node.els = [];
    }
    else if (typeof target === 'string' || typeof target === 'number') {
        const targetElement = queryElement(target);
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (targetElement && els.includes(targetElement)) {
            const mountCount = getCountOfElement(targetElement);
            if (!count || mountCount <= 1) {
                removeElement(targetElement);
                node.els = els.filter(el => el !== targetElement);
            }
            else {
                setCountOfElement(targetElement, mountCount - 1);
            }
        }
    }
    else {
        const mountCount = getCountOfElement(target);
        if (!count || mountCount <= 1) {
            setCountOfElement(target, null);
            target.innerHTML = '';
        }
        else {
            setCountOfElement(target, mountCount - 1);
        }
    }
}
function addElementToList(els, target) {
    if (!els.includes(target)) {
        els.push(target);
    }
}
function mount(instance, node, target, props, count) {
    let targetElement = null;
    const els = node.els;
    if (target === undefined) {
        targetElement = document.createElement('style');
        document.head.appendChild(targetElement);
        addElementToList(els, targetElement);
    }
    else if (typeof target === 'string' || typeof target === 'number') {
        targetElement = queryElement(target);
        if (targetElement === null) {
            targetElement = createElement(target);
            document.head.appendChild(targetElement);
            if (count) {
                setCountOfElement(targetElement, 1);
            }
            addElementToList(els, targetElement);
        }
        else {
            if (count) {
                setCountOfElement(targetElement, getCountOfElement(targetElement) + 1);
            }
            addElementToList(els, targetElement);
            return targetElement;
        }
    }
    else {
        targetElement = target;
        const mountCount = getCountOfElement(targetElement);
        if (mountCount > 0) {
            if (count) {
                setCountOfElement(targetElement, mountCount + 1);
            }
            return target;
        }
        else if (count) {
            setCountOfElement(targetElement, 1);
        }
    }
    const style = node.render(props);
    if (targetElement.innerHTML !== style) {
        targetElement.innerHTML = style;
    }
    return targetElement;
}

/** render wrapper */
function _r(props) {
    return render(this, this.instance, props);
}
const inNode = typeof window === 'undefined';
/** wrapped mount */
function _wm(options) {
    /* istanbul ignore next */
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (inNode)
        return null;
    // eslint-disable-next-line
    const target = options && options.target;
    if (target === null)
        return null;
    // eslint-disable-next-line
    const props = options && options.props;
    // eslint-disable-next-line
    const count = !((options && options.count) === false);
    const targetElement = mount(this.instance, this, target, props, count);
    return targetElement;
}
/** wrapped unmount */
function _wu(options) {
    /* istanbul ignore next */
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (inNode)
        return;
    // eslint-disable-next-line
    const target = options && options.target;
    // eslint-disable-next-line
    const delay = (options && options.delay) || 0;
    // eslint-disable-next-line
    const count = !((options && options.count) === false);
    if (target === null)
        return;
    if (delay === 0)
        unmount(this.instance, this, target, count);
    else {
        setTimeout(() => unmount(this.instance, this, target, count), delay);
    }
}
/** create CNode */
const _cc = function (instance, $, props, children) {
    return {
        instance,
        $,
        props,
        children,
        els: [],
        render: _r,
        mount: _wm,
        unmount: _wu
    };
};
const c = function (instance, $, props, children) {
    if (Array.isArray($)) {
        return _cc(instance, { $: null }, null, $);
    }
    if (Array.isArray(props)) {
        return _cc(instance, $, null, props);
    }
    else if (Array.isArray(children)) {
        return _cc(instance, $, props, children);
    }
    else {
        return _cc(instance, $, props, null);
    }
};

function CSSRender(config = {
    preserveEmptyBlock: false
}) {
    const cssr = {
        c: ((...args) => c(cssr, ...args)),
        use: (plugin, ...args) => plugin.install(cssr, ...args),
        context: {},
        config
    };
    return cssr;
}

export default CSSRender;
export { CSSRender };
